#' Differential Analysis of sRNA clusters or mRNA using `DESeq2` or `edgeR`
#'
#' @description This function allows you to compute the differential analysis
#' of sRNA dicer-derived clusters or mRNA. The function allows the choice 
#' between analysis with `DESeq2` or `edgeR`.
#'
#' @param data data.frame; originally generated by [mobileRNA::RNAimport()].
#'
#' @param group character; vector contains the condition
#' (ie. treatment or control) of each sample in the order samples are shown in 
#' the `data` object from left to right.
#'
#' @param method character; method to undertaken differential analysis, choose 
#' from methods of either [DESeq2::DESeq] or [edgeR::edgeR].
#' Must be stated as either "DESeq2" or "edgeR" in the function.
#'
#' @param dispersionValue numeric; manual setting of dispersion value
#' which is recommended for analysis in experiments without biological
#' replicates when utilising the [edgeR::edgeR] method.
#'
#'
#' @return Undertakes differential analysis, based on a specified method, and
#' adds the results to the supplied dataframe. This includes:
#' * Count mean
#' * Log fold change
#' * p-value
#' * Adjusted p-value
#' * Log counts per million (CPM/RPM)
#' * Comparison order 
#'
#' @details The analysis allows the users to choose the method which best suits
#' their data. Notably, `DESeq2` cannot compute the analysis when there only
#' one replicate per condition, but, `edgeR` can. Simply set a suitable
#' dispersion value, based on similar data, to use this feature. The dispersion
#' value is other wise known as the common Biological squared coefficient
#' of variation. A typical dispersion value is 0.4 for human data sets, 0.1 for
#' data on genetically identical model organisms or 0.01 for technical
#' replicate. See the User’s Guide for the `edgeR` package for more details,
#' [edgeR::edgeR].
#'
#' @examples
#'# load data 
#' data("sRNA_data")
#' 
#' controls <- c("selfgraft_1", "selfgraft_2", "selfgraft_3")
#' 
#' reorder_df <- RNAreorder(sRNA_data, controls)
#' 
#'# sample conditions.
#'groups <- c("Selfgraft", "Selfgraft", "Selfgraft", "Heterograft", "Heterograft", "Heterograft")
#'
#'
#'## Differential analysis: DEseq2 method
#'sRNA_DESeq2 <- RNAdifferentialAnalysis(data = reorder_df,
#'                              group = groups,
#'                              method = "DESeq2" )
#'
#'
#'## Differential analysis: edgeR method
#'sRNA_edgeR <- RNAdifferentialAnalysis(data = reorder_df,
#'                             group = groups,
#'                             method = "edgeR" )
#' @export
#' @importFrom DESeq2 results
#' @importFrom DESeq2 DESeq
#' @importFrom stats relevel
#' @importFrom edgeR exactTest
#' @importFrom stats p.adjust
#' @importFrom dplyr select
#' @importFrom tidyselect starts_with
#' @importFrom dplyr mutate_at
#' @importFrom tidyr replace_na
#' @importFrom DESeq2 DESeqDataSetFromMatrix
#' @importFrom DESeq2 estimateSizeFactors
#' @importFrom edgeR DGEList
#' @importFrom edgeR calcNormFactors
#' @importFrom edgeR estimateDisp

RNAdifferentialAnalysis <- function(data, group, 
                                    method = c("edgeR", "DESeq2"),
                                    dispersionValue = NULL){
  if (base::missing(data) || !base::inherits(data,  "data.frame")) {
    stop("Please specify a data frame")
  }
  if (base::missing(method) || !method %in% c("edgeR", "DESeq2")) {
    stop("Please specify analysis method", "(\"edgeR\", or \"DESeq2\")")
  }
  if ( is.null(group) || !base::inherits(group, "character")) {
    stop("group must be an vector of characters to specify the treatment
         and control conditions for each replicate")
  }
  counts <- data %>% dplyr::select(tidyselect::starts_with("Count"))
  method <- base::match.arg(method)
  if(method == "edgeR"){
    message("Undertaking differential analysis with the edgeR method..")
    res <-  .edgeR_normalise(counts, group)
    if(is.null(dispersionValue)){
      mean <- rowMeans(res$counts)
      groupConditions <- unique(group)
      comp <- edgeR::exactTest(res, pair= groupConditions)
      comp <- base::cbind(comp$table, padj=stats::p.adjust(comp$table$PValue,
                                                           method="BH"))
      comp$CountMean <- mean
    } else {
      comp <- edgeR::exactTest(res, dispersion=dispersionValue^2)
      comp <- base::cbind(comp$table, padj=stats::p.adjust(comp$table$PValue,
                                                           method="BH"))
    }
    data$CountMean <- comp$CountMean
    data$log2FoldChange <- comp$logFC
    data$pvalue <- comp$PValue
    data$padjusted <- comp$padj
    data$logCPM <- comp$logCPM
  } else
    if (method == "DESeq2"){
      message("Undertaking differential analysis with the DESeq2 method..")
      res <- .DESeq_normalise(counts, group)
      comp <- DESeq2::DESeq(res)
      comp <- DESeq2::results(comp)
      data$baseMean <- comp[,"baseMean"]
      data$log2FoldChange <- comp[,"log2FoldChange"]
      data$pvalue <- comp[,"pvalue"]
      data$padjusted <- comp[,"padj"]
    }
  res.df <- data %>%
    dplyr::mutate_at(c('log2FoldChange','padjusted', 'pvalue'),
                     ~tidyr::replace_na(.,0))
  comparision.order <- paste(unique(groups),  collapse = " Vs ")
  res.df$comparision.order <- comparision.order
    
  return(res.df)
}
